#!/usr/bin/env python

import urllib2
import urlparse
from xml.dom import minidom
from HTMLParser import HTMLParser
import sys, os
from optparse import OptionParser
import urllib
import datetime

admin_url = 'http://%s/admin/advanced'
post_urls = ['http://%s/admin/asipura.spa', 'http://%s/admin/linksys.pap']
reboot_url = 'http://%s/admin/reboot'
realms = ['spa admin', 'pap admin']
prog = 'ataconf'

def stderr(msg):
    print >> sys.stderr, msg

def inform(msg):
    if options.noisy: print msg

def iif(cond, val1, val2):
    if cond: return val1
    return val2

def getConfigurationFromDevice(url, username, pw, realm):
    authhandler = urllib2.HTTPDigestAuthHandler()
    authhandler.add_password(realm, url, username, pw)
    opener = urllib2.build_opener(authhandler)
    urllib2.install_opener(opener)
    pagehandle = urllib2.urlopen(url)
    return pagehandle.read()

def writeConfigurationToDevice(url, username, pw, realm, configValues):
    authhandler = urllib2.HTTPDigestAuthHandler()
    authhandler.add_password(realm, url, username, pw)
    opener = urllib2.build_opener(authhandler)
    urllib2.install_opener(opener)
    pagehandle = urllib2.urlopen(url, configValues)
    return pagehandle.read()

class MyHTMLParser(HTMLParser):
    """
    Parse the field names, numbers and values out of the ATA's HTML file. 
    This is fairly straightforward because the code looks like:

    <tr bgcolor="#d3d3d3"><td>Enable Web Server:<td><select class="inputw" name="31215"><option selected value="1">yes</option><option  value="0">no</option></select><td>Web Server Port:<td><input class="inputw" size="20" name="25710" value="80" maxlength=2047>

    So we look for table cells and the following input/select fields.
    """

    def __init__(self):
        HTMLParser.__init__(self)
        self._startedCell = False
        self._startedOption = False
        self._fieldNameToNum = {}
        self._configValues = []
        self._possibleValues = {}
        self._select = None
        self._configValuesByTab = {}

    def findAttr(self, attrs, attr):
        for a, val in attrs:
            if a == attr:
                return val
        return None

    def hasAttr(self, attrs, attr):
        for a, val in attrs:
            if a == attr:
                return True
        return False

    def fixFieldName(self, fieldname):
        # replace characters with underscores to ensure the resulting field name is 
        # valid as a python identifier
        fieldname = fieldname.lower()
        fieldname = fieldname.replace(' ', '_')
        fieldname = fieldname.replace('-', '_')
        fieldname = fieldname.replace('(', '_')
        fieldname = fieldname.replace(')', '_')
        fieldname = fieldname.replace('/', '_')
        return fieldname

    def uniqueFieldName(self, fieldname):
        name = self._tab + '_' + self.fixFieldName(fieldname)
        assert not self._fieldNameToNum.has_key(fieldname), '%s, %s' % (fieldname, attrs)
        return name

    def handle_starttag(self, tag, attrs):
        if tag == 'td':
            self._startedCell = True
        else:
            self._startedCell = False
        if tag == 'div':
            if self.findAttr(attrs, 'class') == 'tab-page':
                self._tab = self.fixFieldName(self.findAttr(attrs, 'id'))
                self._configValuesByTab[self._tab] = []
        if tag == 'input':
            if self.findAttr(attrs, 'type') in ['submit', 'reset']:
                return  # don't do anything
            fieldname = self.uniqueFieldName(self._fieldName)
            self._fieldNameToNum[fieldname] = self.findAttr(attrs, 'name')
            value = self.findAttr(attrs, 'value')
            self._configValuesByTab[self._tab].append((fieldname, value))
        elif tag == 'select':
            fieldname = self.uniqueFieldName(self._fieldName)
            self._fieldNameToNum[fieldname] = self.findAttr(attrs, 'name')
            self._select = fieldname
            self._options = []
            self._selectedOptionValue = None
        elif tag == 'option':
            self._startedOption = True
            value = self.findAttr(attrs, 'value')
            self._curOptionValue = value
            if self.hasAttr(attrs, 'selected'):
                self._selectedOptionValue = value

    def handle_endtag(self, tag):
        self._startedCell = self._startedOption = False
        if tag == 'select':
            self._configValuesByTab[self._tab].append((self._select, self._selectedOptionValue))
            self._possibleValues[self._select] = self._options
        elif tag == 'option':
            self._options.append((self._curOptionValue, self._curOptionDesc))

    def handle_data(self, data):
        if self._startedCell and data[-1] == ':':
            self._fieldName = data[:-1]
        elif self._startedOption:
            self._curOptionDesc = data

    def fieldNameMap(self):
        return self._fieldNameToNum.copy()

    def configValues(self):
        values = []
        for tab, configValues in self._configValuesByTab.items():
            values.extend(configValues)
        return values

    def configValuesByTab(self):
        return self._configValuesByTab.copy()

    def possibleValues(self):
        return self._possibleValues.copy()


def configValuesToFile(configValuesByTab, possibleValues, header):
    lines = [header]
    for tab, configValues in configValuesByTab.items():
        if configValues:
            lines.append('''
#
# %s
#       ''' % tab)
        for fieldname, value in configValues:
            if possibleValues.has_key(fieldname):
                valuestring = ', '.join(map(lambda opt: "'%s'" % opt[1], possibleValues[fieldname]))
                comment = '# possible values: %s\n' % (valuestring)

                # find the description for the current value
                for val, desc in possibleValues[fieldname]:
                    if value == val:
                        value = desc
                        break
            else:
                comment = ''
            line = "%s%s = '%s'" % (comment, fieldname, value)
            lines.append(line)
    return '\n\n'.join(lines) + '\n'

def configFileToDict(input_file, fieldNameMap, possibleValues):
    """
    Read configuration file into a dict.
    """
    config = {}
    if input_file == '-':
        configlines = sys.stdin.readlines()
    elif input_file.startswith('-') and input_file.find('=') != -1:
        configlines = input_file[1:].splitlines()
    else:
        configlines = open(input_file).readlines()    
    for line in configlines:
        if not line or line.startswith('#'):
            continue
        if line.startswith('!'):
            inform(line[1:].strip())
            continue
        if line.find('=') == -1:
            continue
        key, value = line.split('=', 1)
        key = key.strip()
        value = value.strip()
        if (value[0] == "'" and value[-1] == "'") or (value[0] == '"' and value[-1] == '"'):
            value = value[1:-1] # strip quote marks from values
        config[key] = value
        
    mappedConfigValues = {}
    for key, value in config.items():
        if key == '__builtins__':
            continue
        if key.find('password') and value.startswith('****'):
            inform("Ignoring setting: %s = '%s'" % (key, value))
            continue
        if possibleValues.has_key(key):
            for val, desc in possibleValues[key]:
                if value == desc:
                    value = val
        newkey = fieldNameMap[key]
        mappedConfigValues[newkey] = value
    return mappedConfigValues

def parseHost(host):
    i = host.rfind('@')
    if i != -1:
        usernamePassword, hostPort = host[:i], host[i+1:]
        i = usernamePassword.find(':')
        if i != -1:
            username, password = usernamePassword[:i], usernamePassword[i+1:]
    else:
        username = password = ''
        hostPort = host
    i = hostPort.find(':')
    if i != -1:
        host, port = hostPort[:i], hostPort[i+1:]
    else:
        host = hostPort
        port = ''
    return username, password, host, port

def main():
    usage = '''%prog [-nq] <src> [<dest>]

    Saves or restores a configuration from a Linksys or Sipura ATA.

    -n, --noboot   Do not reboot the ATA after configuring it.
    -q, --quiet    Do not print informational messages.

    src      Source of configuration, either a file (- means stdin)
             or the URL of an ATA.  May also be a key-value pair.
    dest     Destination of configuration, a file (- means stdout)
             or the URL of an ATA (if omitted, stdout is used)

For more information, invoke %prog --help.'''

    help = '''%prog [-nq] <src> [<dest>]

    %prog can read the active configuration from a Sipura SPA or Linksys 
    PAP2 analog telephone adapter (ATA) and store it in a text file, or can 
    take an existing text file and write that configuration to the ATA.
    Configuration options are stored as key-value pairs.

    One of <src> or <dest> should be a local file path, and the other 
    should be a URL to your ATA with the username and password embedded.
    If <src> or <dest> is a hyphen (-), standard input or output is used.
    <src> can also be a key-value pair (i.e., key = value).

    Options:

    -n, --noboot   Do not reboot the ATA after configuring it.
    -q, --quiet    Do not print informational messages.

    Examples:

    Write ATA config to standard output:
        %prog http://admin:mypassword@myata/
    or  %prog http://admin:mypassword@myata/ -

    Copy ATA config into a config file:
        %prog http://admin:mypassword@myata/ ata.config

    Or use an explicit IP if necessary:
        %prog http://admin:mypassword@192.168.0.5/ ata.config

    Copy a config file to the ATA device (no reboot afterward):
        %prog --noboot ata.config http://admin:mypassword@192.168.0.5/

    Copy config options from standard input to the ATA device:
        %prog - http://admin:mypassword@myata/

    Set a single option (key/value pair):
        %prog key = value http://admin:mypassword@myata/
    or  %prog key value http://admin:mypassword@myata/

    Notes:
        When reading a config from an ATA, this script unfortunately
        cannot retrieve the passwords.  They appear as stars ("****").  
        It is however safe to write such a configuration back to the 
        device; such passwords will be ignored, and the passwords in the 
        device will be left unchanged.  If you edit a config file to
        contain a (non-star) password, however, it will be changed.

        It is not necessary that a config file contain key-value pairs
        for every option supported by the ATA. Any options not present 
        will not be changed.

        Lines beginning with # in a config file are taken as comments and
        ignored.  Lines beginning with ! are printed to standard out when
        the config file is restored to an ATA.

        Single and double quotes are permitted around option values (single
        quotes are used in generated files) but may be omitted.  When using
        a key-value pair in the command, it is good practice to use single
        quotes around the value to keep the shell from munging it.

        This program has been tested with the Sipura SPA 3000 (firmware
        versions 2.0.13g and 3.1.3), and with the Linksys PAP2 (firmware 
        3.1.9 and 5.1.5).  Your mileage may vary.'''

    optparser = OptionParser(usage=usage, version="%prog 0.3")
    optparser.remove_option('--help')
    optparser.remove_option('--version')    
    optparser.add_option('-n', '--noboot', action='store_false', default=True, dest='reboot')
    optparser.add_option('-q', '--quiet', action='store_false', default=True, dest='noisy')
    optparser.add_option('-h', '--help', action='store_true', default=False, dest='help')
    optparser.add_option('-v', '--version', action='store_true', default=False, dest='version')
    optparser.disable_interspersed_args()
    global options; (options, args) = optparser.parse_args()

    if options.version: optparser.print_version(); return 2
    if options.help: optparser.set_usage(help); optparser.print_usage(); return 0

    if len(args) == 0:
        optparser.print_usage()
        return 2
    urls = 0
    for arg in args:
        if arg.startswith('http://'):
            urls += 1
    if urls == 0:
        optparser.error('Either the source or the destination must be a URL.')
    elif urls >= 2:
        optparser.error('Only the source OR the destination should be a URL; the other should be a local file.')
    if args[0].startswith('http://'):
        url = args[0]
        action = 'read-config'
        if len(args) >= 2:
            options.output_file = args[1]
        else:
            options.output_file = '-'   # output to stdout
    else:
        if len(args) < 2:
            optparser.error('If the source is a file, you must specify the destination (the URL to an ATA).')
        url = args[1]
        action = 'write-config'
        options.input_file = args[0]
        if len(args) == 3 and args[2].startswith('http://'):
            url = args[2]
            options.input_file = '-%s = %s' % (args[0], args[1])
        elif len(args) == 4 and args[1] == '=' and args[3].startswith('http://'):
            url = args[3]
            options.input_file = '-%s = %s' % (args[0], args[2])
        elif len(args) != 2:
            optparser.error('Too many arguments.')

    # break apart url
    scheme, host, uri = urlparse.urlparse(url)[:3]
    username, password, host, port = parseHost(host)
    if port:
        hostPort = '%s:%s' % (host, port)
    else:
        hostPort = host

    if not username or not password:
        optparser.error('You need to specify the username and password for the ATA.')

    # figure out realm and retrieve configuration page
    url = admin_url % hostPort
    foundrealm = False
    for realm in realms:
        try:
            html = getConfigurationFromDevice(url, username, password, realm)
        except urllib2.HTTPError, e:
            if e.code != 401:
                stderr('HTTP Error %d (%s) occurred when communicating with the ATA.' % (e.code, e.msg))
                return e.code
        except urllib2.URLError, e:
            if hasattr(e, 'reason'):
                stderr('Error %d (%s) occurred when communicating with the ATA.' % (e.reason[0], e.reason[1]))
                if e.reason[0] == 111:
                    stderr('Check that you have the correct hostname and port number.')
                return e.reason[0]
            else:
                stderr(str(e))
                return e
        else:
            foundrealm = True
            break
    if not foundrealm:
        stderr("Couldn't log in to the ATA. Usually this means an invalid password.\nI tried these realms:")
        for realm in realms:
            stderr('\t' + realm)
        return 401

    # parse configuration page
    htmlparser = MyHTMLParser()
    htmlparser.feed(html)

    try:
        if action == 'read-config':
            # the ATA configuration page has already been read and parsed
            if options.output_file != '-':
                out = open(options.output_file, 'w')
            else:
                out = sys.stdout
            header = '! ATA config from %s at %s' % (hostPort, datetime.datetime.now().ctime())
            out.write(configValuesToFile(htmlparser.configValuesByTab(), htmlparser.possibleValues(), header))
            if options.output_file != '-':
                inform('Saved ATA configuration to file %s.' % options.output_file)
        elif action == 'write-config':
            # get the fieldname->number mapping from the parsed config page
            mapping = htmlparser.fieldNameMap()
            possibleValues = htmlparser.possibleValues()

            # read the config file
            params = configFileToDict(options.input_file, mapping, possibleValues)
            
            params = urllib.urlencode(params)
            for post_url in post_urls:
                url = post_url % hostPort
                foundurl = False
                try:
                    writeConfigurationToDevice(url, username, password, realm, params)
                except urllib2.HTTPError, e:
                    if e.code != 404:
                        raise
                else:
                    foundurl = True
                    break
            if not foundurl:
                stderr("Couldn't find correct URL to save configuration to ATA.\nI tried these URLs:")
                for post_url in post_urls:
                    stderr('\t' + post_url)
            else:
                inform('Installed config to %s.  %s.' % (hostPort, iif(options.reboot, 'Rebooting', 'No reboot')))
                if options.reboot:
                    url = reboot_url % hostPort
                    try:
                         getConfigurationFromDevice(url, username, password, realm)
                    except:
                         return 0
    except urllib2.HTTPError, e:
        stderr('HTTP Error %d (%s) occurred when communicating with the ATA.' % (e.code, e.msg))
    except urllib2.URLError, e:
        if hasattr(e, 'reason'):
            stderr('Error %d (%s) occurred when communicating with the ATA.' % (e.reason[0], e.reason[1]))
        else:
            stderr(str(e))
    except IOError, e:
        try:
            (code, message) = e
        except:
            code = 0
            message = e
        if code != 32: # suppress message for broken pipe
            stderr('I/O Error: %s (%d)' % (message, code))
        return code            
    return 0

if __name__ == '__main__':
    sys.exit(main())
